
from unittest import mock

import pytest
import warnings

import zulu


parametrize = pytest.mark.parametrize


@parametrize('api_func, target_func', [
    (zulu.now, 'zulu.zulu.Zulu.now'),
    (zulu.parse, 'zulu.zulu.Zulu.parse'),
    (zulu.range, 'zulu.zulu.Zulu.range'),
    (zulu.span_range, 'zulu.zulu.Zulu.span_range'),
    (zulu.parse_delta, 'zulu.delta.Delta.parse')
])
def test_api_mapping(api_func, target_func):
    with mock.patch(target_func) as mocked_target_func:
        api_func()

    assert mocked_target_func.called




from datetime import timedelta
import pickle

import pytest

from zulu import ParseError, Delta


parametrize = pytest.mark.parametrize


@parametrize('obj,expected', [
    (timedelta(seconds=60), Delta(seconds=60)),
    (60, Delta(seconds=60)),
    ('32m', Delta(minutes=32)),
    ('2h32m', Delta(hours=2, minutes=32)),
    ('3d2h32m', Delta(days=3, hours=2, minutes=32)),
    ('1w3d2h32m', Delta(weeks=1, days=3, hours=2, minutes=32)),
    ('1w 3d 2h 32m', Delta(weeks=1, days=3, hours=2, minutes=32)),
    ('1 w 3 d 2 h 32 m', Delta(weeks=1, days=3, hours=2, minutes=32)),
    ('4:13', Delta(minutes=4, seconds=13)),
    ('4:13:02', Delta(hours=4, minutes=13, seconds=2)),
    ('4:13:02.266', Delta(hours=4,
                          minutes=13,
                          seconds=2,
                          microseconds=266000)),
    ('2:04:13:02.266', Delta(days=2,
                             hours=4,
                             minutes=13,
                             seconds=2,
                             microseconds=266000)),
    ('2 days, 4:13:02', Delta(days=2, hours=4, minutes=13, seconds=2)),
    ('2 days, 4:13:02.266', Delta(days=2,
                                  hours=4,
                                  minutes=13,
                                  seconds=2,
                                  microseconds=266000)),
    ('5hr34m56s', Delta(hours=5, minutes=34, seconds=56)),
    ('5 hours, 34 minutes, 56 seconds', Delta(hours=5,
                                              minutes=34,
                                              seconds=56)),
    ('5 hrs, 34 mins, 56 secs', Delta(hours=5, minutes=34, seconds=56)),
    ('2 days, 5 hours, 34 minutes, 56 seconds', Delta(days=2,
                                                      hours=5,
                                                      minutes=34,
                                                      seconds=56)),
    ('1.2 m', Delta(seconds=72)),
    ('1.2 min', Delta(seconds=72)),
    ('1.2 mins', Delta(seconds=72)),
    ('1.2 minute', Delta(seconds=72)),
    ('1.2 minutes', Delta(seconds=72)),
    ('172 hours', Delta(days=7, seconds=14400)),
    ('172 hr', Delta(days=7, seconds=14400)),
    ('172 h', Delta(days=7, seconds=14400)),
    ('172 hrs', Delta(days=7, seconds=14400)),
    ('172 hour', Delta(days=7, seconds=14400)),
    ('1.24 days', Delta(days=1, seconds=20736)),
    ('5 d', Delta(days=5)),
    ('5 day', Delta(days=5)),
    ('5 days', Delta(days=5)),
    ('5.6 wk', Delta(days=39, seconds=17280)),
    ('5.6 week', Delta(days=39, seconds=17280)),
    ('5.6 weeks', Delta(days=39, seconds=17280)),
])
def test_delta_parse(obj, expected):
    assert Delta.parse(obj) == expected


def test_delta_as_float():
    secs = 10.1234
    delta = Delta(seconds=secs)
    assert float(delta) == secs


def test_delta_as_int():
    secs = 10.1234
    delta = Delta(seconds=secs)
    assert int(delta) == int(secs)


@parametrize('obj,exception', [
    ({}, TypeError),
    ('', ParseError),
    ('a', ParseError),
])
def test_delta_parse_invalid(obj, exception):
    with pytest.raises(exception):
        Delta.parse(obj)


@parametrize('delta,opts,expected', [
    (Delta(seconds=5), {}, '5 seconds'),
    (Delta(seconds=5), {'format': 'short'}, '5 sec'),
    (Delta(seconds=5), {'format': 'narrow'}, '5s'),
    (Delta(seconds=5), {'add_direction': True}, 'in 5 seconds'),
    (Delta(seconds=-5), {'add_direction': True}, '5 seconds ago'),
    (Delta(days=1, hours=10), {}, '1 day'),
    (Delta(days=1), {'granularity': 'year'}, '1 year'),
    (Delta(minutes=15), {'threshold': 0.15}, '0 hours'),
    (Delta(hours=23), {'threshold': 0.85}, '1 day'),
    (Delta(hours=23), {'threshold': 1.00}, '23 hours'),
    (Delta(hours=23), {'threshold': 1.00, 'locale': 'de'}, '23 Stunden'),
])
def test_delta_format(delta, opts, expected):
    assert delta.format(**opts) == expected


def test_delta_format_default_locale(monkeypatch):
    for var in ('LANGUAGE', 'LC_ALL', 'LC_CTYPE', 'LANG'):
        monkeypatch.setenv(var, '')

    assert Delta(seconds=5).format() == '5 seconds'


@parametrize('opts,exception', [
    ({'granularity': 'invalid'}, ValueError),
    ({'format': 'invalid'}, ValueError),
])
def test_delta_format_exception(opts, exception):
    with pytest.raises(exception):
        Delta(1).format(**opts)


def test_delta_math_operations_return_type():
    delta = Delta(days=1, hours=1, minutes=1, seconds=1, microseconds=1)

    assert isinstance(delta + delta, Delta)
    assert isinstance(delta - delta, Delta)
    assert isinstance(delta * 1, Delta)
    assert isinstance(delta / 1, Delta)
    assert isinstance(delta // 1, Delta)
    assert isinstance(abs(delta), Delta)
    assert isinstance(+delta, Delta)
    assert isinstance(-delta, Delta)
    assert isinstance(delta % delta, Delta)
    assert isinstance(divmod(delta, delta)[1], Delta)


def test_delta_pickle():
    delta = Delta(hours=1)
    assert pickle.loads(pickle.dumps(delta)) == delta



"""The zulu module.
"""

from .__version__ import __version__

from .api import (
    create,
    now,
    parse,
    parse_delta,
    range,
    span_range
)
from .zulu import Zulu
from .parser import ParseError, ISO8601, TIMESTAMP
from .delta import Delta



"""The api module.
"""

from .zulu import Zulu
from .delta import Delta


def create(*args, **kargs):
    """Factory function to create a new :class:`.Zulu` datetime using the same
    arguments as the class.

    .. seealso:: See :class:`.Zulu` for class signature details.
    """
    return Zulu(*args, **kargs)


def now():
    """Alias to :meth:`.Zulu.now`."""
    return Zulu.now()


def parse(*args, **kargs):
    """Alias to :meth:`.Zulu.parse`.

    .. seealso:: See :meth:`.Zulu.parse` for function signature details.
    """
    return Zulu.parse(*args, **kargs)


def range(*args, **kargs):
    """Alias to :meth:`.Zulu.range`.

    .. seealso:: See :meth:`.Zulu.range` for function signature details.
    """
    return Zulu.range(*args, **kargs)


def span_range(*args, **kargs):
    """Alias to :meth:`.Zulu.span_range`.

    .. seealso:: See :meth:`.Zulu.span_range` for function signature details.
    """
    return Zulu.span_range(*args, **kargs)


def parse_delta(*args, **kargs):
    """Alias to :meth:`.Delta.parse`.

    .. seealso:: See :meth:`.Delta.parse` for function signature details.
    """
    return Delta.parse(*args, **kargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


package_dir = 'src'
packages = find_packages(package_dir)

version = {}
with open('{}/{}/__version__.py'.format(package_dir, packages[0])) as f:
    exec(f.read(), version)


setup(
    version=version['__version__'],
    package_dir={'': package_dir},
    packages=packages,
    include_package_data=True
)



"""The parser module.
"""

from __future__ import absolute_import

from functools import partial
from itertools import groupby
from datetime import datetime, timedelta

from babel.dates import (
    LC_TIME,
    format_timedelta as _format_timedelta,
    format_datetime as _format_datetime
)
from dateutil.tz import gettz, tzlocal, tzutc
import iso8601
import pytimeparse

from .helpers import number_types


UTC = tzutc()
EPOCH = datetime(1970, 1, 1, tzinfo=UTC)

ISO8601 = 'ISO8601'
TIMESTAMP = 'timestamp'
DEFAULT_PARSE_DATETIME_FORMATS = (ISO8601, TIMESTAMP)


# Subset of Unicode date field patterns from:
# https://www.unicode.org/reports/tr35/tr35-dates.html#Date_Field_Symbol_Table
# that are supported as an alternative to Python's strptime/strftime
# directives. This mapping is used to convert the Unicode pattern (the dict
# keys) to the best matched strptime directive (the values). For values that
# are tuples, the first item is the directive used for string parsing while the
# second item is the directive used for string formatting. The second items are
# platform dependent and may not work on all systems.
DATE_PATTERN_TO_DIRECTIVE = {
    'YYYY': '%Y',    # Year with century
    'YY': '%y',      # Year without century
    'MMMM': '%B',    # Month's full name
    'MMM': '%b',     # Month's abbreviated name
    'MM': '%m',      # Month padded
    'M': '%m',       # Month not padded
    'DDD': '%j',     # Day of the year padded
    'DD': '%j',      # Day of the year padded
    'D': '%j',       # Day of the year not padded
    'dd': '%d',      # Day of the month padded
    'd': '%d',       # Day of the month not padded
    'EEEE': '%A',    # Weekday's full name
    'EEE': '%a',     # Weekday's abbreviated name
    'EE': '%a',      # Weekday's abbreviated name
    'E': '%a',       # Weekday's abbreviated name
    'eee': '%a',     # Weekday's abbreviated name
    'ee': '%w',      # Weekday as decimal padded
    'e': '%w',       # Weekday as decimal not padded
    'HH': '%H',      # Hour-24 padded
    'H': '%H',       # Hour-24 not padded
    'hh': '%I',      # Hour-12 padded
    'h': '%I',       # Hour-12 not padded
    'mm': '%M',      # Minute padded
    'm': '%M',       # Minute not padded
    'ss': '%S',      # Second padded
    's': '%S',       # Second not padded
    'SSSSSS': '%f',  # Microsecond padded
    'SSSSS': '%f',   # Microsecond padded
    'SSSS': '%f',    # Microsecond padded
    'SSS': '%f',     # Microsecond padded
    'SS': '%f',      # Microsecond padded
    'S': '%f',       # Microsecond not padded
    'a': '%p',       # am or pm
    'z': '%z',       # UTC offset without separator
    'Z': '%z',       # UTC offset without separator
}

TIMEDELTA_GRANULARITIES = ('second',
                           'minute',
                           'hour',
                           'day',
                           'week',
                           'month',
                           'year')

TIMEDELTA_FORMATS = ('long', 'short', 'narrow')


class ParseError(Exception):
    """Exception raised when an object cannot be parsed as a datetime."""
    pass


def parse_datetime(obj, formats=None, default_tz=None):
    """Attempt to parse `obj` as a ``datetime`` using  a list of `formats`. If
    no timezone information is found in `obj` and `default_tz` is set, then
    the naive datetime object will be shifted to the default timezone.

    Args:
        obj (str|datetime): Object to parse.
        formats (list, optional): List of string formats to use when parsing.
            Defaults to ``['ISO8601', 'X']``.
        default_tz (None|str|tzinfo, optional): Default timezone to use when
            parsed datetime object does not contain a timezone. Defaults to
            ``UTC``.

    Returns:
        datetime

    Raises:
        ValueError: When `default_tz` is an unrecognized timezone.
        ParseError: When `obj` can't be parsed as a datetime.
    """
    if default_tz is None:
        default_tz = UTC

    if not is_valid_timezone(default_tz):
        raise ValueError('Unrecognized timezone: {0}'.format(default_tz))

    if is_valid_datetime(obj):
        return obj

    if formats is None:
        formats = DEFAULT_PARSE_DATETIME_FORMATS
    elif not isinstance(formats, (list, tuple)):
        formats = [formats]

    dt = _parse_datetime_formats(obj, formats)

    if dt.tzinfo is None and default_tz is not None:
        dt = dt.replace(tzinfo=get_timezone(default_tz))

    if not has_valid_timezone(dt):  # pragma: no cover
        raise ParseError('Timezone offset must be strictly between -24/+24 '
                         'hours')

    return dt


def _parse_datetime_formats(obj, formats):
    """Parse `obj` as datetime using list of `formats`."""
    dt = None
    errors = {}

    for format in formats:
        try:
            dt = _parse_datetime_format(obj, format)
        except Exception as exc:
            errors[format] = str(exc)
            dt = None
        else:
            break

    if dt is None:
        err = ', '.join('"{0}" ({1})'.format(format, errors[format])
                        for format in formats)
        raise ParseError('Value "{0}" does not match any format in [{1}]'
                         .format(obj, err))

    return dt


def _parse_datetime_format(obj, format):
    """Parse `obj` as datetime using `format`."""
    if format.upper() == ISO8601:
        return iso8601.parse_date(obj, default_timezone=None)
    elif format.lower() == TIMESTAMP:
        return datetime.fromtimestamp(obj, UTC)
    else:
        if '%' not in format:
            format = _date_pattern_to_directive(format)
        return datetime.strptime(obj, format)


def format_datetime(dt, format=None, tz=None, locale=LC_TIME):
    """Return string formatted datetime, `dt`, using format directives or
    pattern in `format`. If timezone, `tz`, is supplied, the datetime will be
    shifted to that timezone before being formatted.

    Args:
        dt (datetime): A datetime instance.
        format (str, optional): Datetime format string. Defaults to ``None``
            which uses ISO-8601 format.
        tz (None|str|tzinfo, optional): Timezone to shift `dt` to before
            formatting.
        locale (str|Locale, optional): A ``Locale`` object or locale
            identifer. Defaults to system default.

    Returns:
        str
    """
    if not isinstance(dt, datetime):
        raise TypeError("zulu.parser.format()'s first argument must be a "
                        "datetime, not {0}"
                        .format(type(dt).__name__))  # pragma: no cover

    if format is not None and not isinstance(format, str):
        raise TypeError("zulu.parser.format()'s second argument must be a "
                        "string or None, not {0}"
                        .format(type(format).__name__))  # pragma: no cover

    if not is_valid_timezone(tz):  # pragma: no cover
        raise ValueError('Unrecognized timezone: {0}'.format(tz))

    if format is None:
        format = ISO8601

    if tz is not None:
        dt = dt.astimezone(tz)

    if format == ISO8601:
        return dt.isoformat()
    elif '%' in format:
        return dt.strftime(format)
    else:
        return _format_datetime(dt, format, locale=locale)


def _date_pattern_to_directive(format):
    """Convert date pattern format to strptime/strftime directives."""
    return ''.join(DATE_PATTERN_TO_DIRECTIVE.get(token, token)
                   for token in _tokenize_date_pattern(format))


def _tokenize_date_pattern(format):
    """Return list of date pattern tokens.

    This groups tokens by repeating characters so that each set of repeating
    characters is a list item (e.g. ``'YY-MM-dd'`` becomes
    ``['YY', '-', 'MM', '-', 'dd']``).
    """
    return [''.join(group) for key, group in groupby(format)]


def parse_timedelta(obj):
    """Attempt to parse `obj` as a ``timedelta`` from a string formatted
    duration.

    Args:
        obj (str|number|timedelta): Object to parse.

    Returns:
        timedelta

    Raises:
        TypeError: When `obj` is not a string or timedelta.
        ParseError: When `obj` can't be parsed as a timedelta.
    """
    if isinstance(obj, timedelta):
        return obj

    is_string = isinstance(obj, str)
    is_number = isinstance(obj, number_types)

    if not is_string and not is_number:
        raise TypeError('Expected string or number type, not {0}'
                        .format(type(obj).__name__))

    if is_string:
        seconds = pytimeparse.parse(obj)

        if seconds is None:
            raise ParseError('Value "{0}" is not a recognized duration format'
                             .format(obj))
    else:
        seconds = obj

    return timedelta(seconds=seconds)


def format_timedelta(delta,
                     format='long',
                     granularity='second',
                     threshold=0.85,
                     add_direction=False,
                     locale=LC_TIME):
    """Return timedelta as a formatted string.

    Args:
        format (str, optional): Can be one of "long", "short", or "narrow".
            Defaults to `'long`'.
        granularity (str, optional): The smallest unit that should be
            displayed. The value can be one of "year", "month", "week",
            "day", "hour", "minute" or "second". Defaults to `'second'`.
        threshold (float, optional): Factor that determines at which point
            the presentation switches to the next higher unit. Defaults to
            `0.85`.
        add_direction (bool, optional): If ``True`` the return value will
            include directional information (e.g. `'1 hour ago'`,
            `'in 1 hour'`). Defaults to ``False``.
        locale (str|Locale, optional): A ``Locale`` object or locale
            identifer. Defaults to system default.

    Returns:
        str
    """
    if granularity not in TIMEDELTA_GRANULARITIES:
        units = ', '.join('"{0}"'.format(unit)
                          for unit in TIMEDELTA_GRANULARITIES)
        raise ValueError('Time delta granularity must be one of {0}, not "{1}"'
                         .format(units, granularity))

    if format not in TIMEDELTA_FORMATS:
        formats = ', '.join('"{0}"'.format(format)
                            for format in TIMEDELTA_FORMATS)
        raise ValueError('Time delta format must be one of {0}, not "{1}"'
                         .format(formats, format))

    return _format_timedelta(delta,
                             granularity=granularity,
                             threshold=threshold,
                             add_direction=add_direction,
                             format=format,
                             locale=locale)


def get_timezone(tz):
    """Coerce `tz` into a `tzinfo` compatible object. If ``tz == 'local'``,
    then the system's local timezone will be used. If `tz` is a string other
    than ``'local'``, it will be passed to ``dateutil.tz.gettz(tz)``.
    Otherwise, `tz` will be returned as-is.
    """
    if tz is None:
        tz = UTC
    elif tz == 'local':
        tz = tzlocal()
    elif isinstance(tz, str):
        tz_string = tz
        tz = gettz(tz)

        if tz is None:
            raise ValueError('Unrecognized timezone string: {0}'
                             .format(tz_string))

    return tz


def get_timestamp(dt):
    """Return timestamp for datetime, `dt`."""
    return (dt - EPOCH).total_seconds()


def is_valid_datetime(obj):
    """Return whether `obj` is an instance of ``datetime`` or contains date and
    time attributes.

    Returns:
        bool
    """
    if isinstance(obj, datetime):
        return True
    else:
        attrs = ('year',
                 'month',
                 'day',
                 'hour',
                 'minute',
                 'second',
                 'microsecond',
                 'tzinfo')
        return all(hasattr(obj, attr) for attr in attrs)


def is_valid_timezone(tz):
    """Return whether `tz` is a valid timezone.

    Returns:
        bool
    """
    try:
        get_timezone(tz)
    except Exception:
        return False
    else:
        return True


def has_valid_timezone(dt):
    """Return whether `dt` has a valid timezone with a UTC offset strictly
    between -24/+24 hours.

    Returns:
        bool
    """
    try:
        dt.astimezone(UTC)
    except Exception:  # pragma: no cover
        return False
    else:
        return True
